/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2009-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl> 
 *
 *  2007-2009
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 *  2003-2007
 *     Martijn Steenbakkers <martijn@nikhef.nl>
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "utils.h"

#include <lcmaps/lcmaps.h>


#ifndef USE_GLOBUS_INTERFACE_TO_LCMAPS /* Make use of the Globus interface to LCMAPS */

#include <lcmaps/lcmaps_account.h>




#ifndef LIBLCMAPS_RETURN_ACCOUNT_FROM_PEM_SO 
    #define LIBLCMAPS_RETURN_ACCOUNT_FROM_PEM_SO "liblcmaps_return_account_from_pem.so"
#endif



int run_lcmaps (char * pem_string, FILE * logfile)
{
    int                     retval        = 0;
    int                     mapcounter    = -1;
    char *                  error         = NULL;

    lcmaps_account_info_t * paccount_info = NULL;

#ifndef LINKED_LCAS_LCMAPS
    void * lcmaps_return_account_from_pem_handle = NULL;

    int (*LcmapsInit)(FILE *);
    int (*LcmapsTerm)();
    int (*Lcmaps_Account_Info_Init)(lcmaps_account_info_t *);
    int (*Lcmaps_Return_Account_From_Pem)(char *, int, lcmaps_account_info_t *);
    /* int (*Lcmaps_Run_With_Pem_And_Return_Account)(char *, char *, int, lcmaps_request_t, int, char **, uid_t *, gid_t **, int *, gid_t **, int *, char **); */
#endif


    /* No PEM string, no nothing */
    if (!pem_string)
    {
        logmsg(LOG_ERR, "Couldn't extract the client certificate credentials.");
        return 1;
    }

    
#ifndef LINKED_LCAS_LCMAPS
    /* lcmaps_return_account_from_pem_handle = dlopen(LIBLCMAPS_RETURN_ACCOUNT_FROM_PEM_SO, RTLD_NOW|RTLD_GLOBAL); */
    lcmaps_return_account_from_pem_handle = dlopen(LIBLCMAPS_RETURN_ACCOUNT_FROM_PEM_SO, RTLD_LAZY|RTLD_GLOBAL);
    if (!lcmaps_return_account_from_pem_handle)
    {
        logmsg(LOG_ERR, "Failed to dynamically load the library for LCMAPS: \"%s\"\n", LIBLCMAPS_RETURN_ACCOUNT_FROM_PEM_SO);
        return 1;
    }
 

    /* Check the symbols */
    LcmapsInit=dlsym(lcmaps_return_account_from_pem_handle,"lcmaps_init");
    if ((error = dlerror()) != NULL)
    {
        logmsg(LOG_ERR, "LCMAPS module not compliant: Symbol \"%s\" not found: %s", "lcmaps_init", error);
        dlclose(lcmaps_return_account_from_pem_handle);
        return 1;
    }

    Lcmaps_Account_Info_Init = dlsym(lcmaps_return_account_from_pem_handle, "lcmaps_account_info_init");
    if ((error = dlerror()) != NULL)
    {
        logmsg(LOG_ERR, "LCMAPS module not compliant: Symbol \"%s\" not found: %s", "lcmaps_account_info_init", error);
        dlclose(lcmaps_return_account_from_pem_handle);
        return 1;
    }

    Lcmaps_Return_Account_From_Pem = dlsym(lcmaps_return_account_from_pem_handle, "lcmaps_return_account_from_pem");
    if ((error = dlerror()) != NULL)
    {
        logmsg(LOG_ERR, "LCMAPS module not compliant: Symbol \"%s\" not found: %s", "lcmaps_return_account_from_pem", error);
        dlclose(lcmaps_return_account_from_pem_handle);
        return 1;
    }

    LcmapsTerm=dlsym(lcmaps_return_account_from_pem_handle,"lcmaps_term");
    if ((error = dlerror()) != NULL)
    {
        logmsg(LOG_ERR, "LCMAPS module not compliant: Symbol \"%s\" not found: %s", "lcmaps_term", error);
        dlclose(lcmaps_return_account_from_pem_handle);
        return 1;
    }
#endif


    /* initialize LCMAPS */
#ifdef LINKED_LCAS_LCMAPS
    retval=lcmaps_init(logfile);
    if (retval)
    {
        logmsg(LOG_ERR, "LCMAPS initialization failure.");
		return 1;
    }
#else
    retval=(*LcmapsInit)(logfile);
    if (retval)
    {
        logmsg(LOG_ERR, "LCMAPS initialization failure.");
        dlclose(lcmaps_return_account_from_pem_handle);
		return 1;
    }
#endif

    paccount_info = malloc (sizeof(lcmaps_account_info_t));
    if (!paccount_info)
    {
        logmsg(LOG_ERR, "Error: Insufficient memory for an lcmaps_account_info_t (%d bytes), aborting\n", sizeof(lcmaps_account_info_t));
#ifdef LINKED_LCAS_LCMAPS
        lcmaps_term();
#else
        (*LcmapsTerm)();
        dlclose(lcmaps_return_account_from_pem_handle);
#endif
        return 1;
    }
        

    /* LCMAPS mapping call */
#ifdef LINKED_LCAS_LCMAPS
    if (lcmaps_account_info_init(paccount_info) != 0) 
    {
        logmsg(LOG_ERR, "Error initializing lcmaps account\n");
        lcmaps_term();
        free(paccount_info);
        return 1;
    }
#else
    if ((*Lcmaps_Account_Info_Init)(paccount_info) != 0) 
    {
        logmsg(LOG_ERR, "Error initializing lcmaps account\n");
        (*LcmapsTerm)();
        dlclose(lcmaps_return_account_from_pem_handle);
        free(paccount_info);
        return 1;
    }
#endif


    /* Return account from PEM */
#ifdef LINKED_LCAS_LCMAPS
    if (lcmaps_return_account_from_pem(pem_string, mapcounter, paccount_info) != 0) 
    {
        logmsg(LOG_WARNING, "Warning: failed mapping. LCMAPS returned: %d\n", retval);
        lcmaps_term();
        free(paccount_info);
        return 1;
    }
#else
    if ((retval = (*Lcmaps_Return_Account_From_Pem)(pem_string, mapcounter, paccount_info)) != 0)
    {
        logmsg(LOG_WARNING, "Warning: failed mapping. LCMAPS returned: %d\n", retval);
        (*LcmapsTerm)();
        dlclose(lcmaps_return_account_from_pem_handle);
        free(paccount_info);
        return 1;
    }
#endif


    /* Mapping is succesful */
#ifdef LINKED_LCAS_LCMAPS
    lcmaps_term();
#else
    (*LcmapsTerm)();
    dlclose(lcmaps_return_account_from_pem_handle);
#endif

    logmsg(LOG_DEBUG, "LCMAPS Terminated");
    free(paccount_info);

    /* All ok */
    logmsg(LOG_NOTICE, "GT4 Authorization & Mapping Interface Succeeded.\n");
    return 0;
}


/***************************************************************************************/
#else /* USE_GLOBUS_INTERFACE_TO_LCMAPS  :  Make use of the Globus interface to LCMAPS */
/***************************************************************************************/


#ifndef LIBLCMAPS_SO
    #define LIBLCMAPS_SO "liblcmaps.so"
#endif

int run_lcmaps(gss_cred_id_t user_cred_handle, char * client_name, FILE * logfile)
{
    /* USER MAPPING */
#if LINKED_LCAS_LCMAPS
        {
            int retval;

            /* Initialize, do user mapping and terminate LCMAPS */
            retval=lcmaps_init(logfile);
            if (retval)
            {
                logmsg(LOG_ERR, "LCMAPS initialization failure.");
                return 1;
            }
#if ALLOW_EMPTY_CREDENTIALS
            retval=lcmaps_run(config->client_name, user_cred_handle, config->lcmaps_request);
#else
            retval=lcmaps_run(user_cred_handle, config->lcmaps_request);
#endif /* ALLOW_EMPTY_CREDENTIALS */
            if (retval)
            {
                logmsg(LOG_WARNING, "Warning: failed mapping. LCMAPS returned: %d\n", retval);
                return 1;
            }
            retval=lcmaps_term();
            if (retval)
            {
                logmsg(LOG_ERR, "LCMAPS Termination failure!");
                return 1;
            }
        }
#else /* LINKED_LCAS_LCMAPS */
        {
            void *lcmaps_handle=NULL;
            char *error;
            int retval;
            int (*LcmapsInit)(FILE *);
            int (*LcmapsTerm)();
#if ALLOW_EMPTY_CREDENTIALS
            int (*LcmapsRun)(char*, gss_cred_id_t, char*);
#else
            int (*LcmapsRun)(gss_cred_id_t, char*);
#endif /* ALLOW_EMPTY_CREDENTIALS */

            lcmaps_handle = dlopen(LIBLCMAPS_SO,RTLD_LAZY|RTLD_GLOBAL);
            if (!lcmaps_handle)
            {
                logmsg(LOG_ERR, "Failed to dynamically load the library for LCMAPS: \"%s\"\n", LIBLCMAPS_SO);
                return 1;
            }
            else
            {
                /* Check the symbols */
                LcmapsInit=dlsym(lcmaps_handle,"lcmaps_init");
                if ((error = dlerror()) != NULL)
                {
                    logmsg(LOG_ERR, "LCMAPS module not compliant: Symbol \"%s\" not found: %s", "lcmaps_init", error);
                    dlclose(lcmaps_handle);
                    return 1;
                }
                LcmapsRun=dlsym(lcmaps_handle,"lcmaps_run");
                if ((error = dlerror()) != NULL)
                {
                    logmsg(LOG_ERR, "LCMAPS module not compliant: Symbol \"%s\" not found: %s", "lcmaps_run", error);
                    dlclose(lcmaps_handle);
                    return 1;
                }
                LcmapsTerm=dlsym(lcmaps_handle,"lcmaps_term");
                if ((error = dlerror()) != NULL)
                {
                    logmsg(LOG_ERR, "LCMAPS module not compliant: Symbol \"%s\" not found: %s", "lcmaps_term", error);
                    dlclose(lcmaps_handle);
                    return 1;
                }

                /* Initialize, do user mapping and terminate LCMAPS */
                retval=(*LcmapsInit)(logfile);
                if (retval)
                {
                    logmsg(LOG_ERR, "LCMAPS initialization failure.");
                    dlclose(lcmaps_handle);
                    return 1;
                }

#if ALLOW_EMPTY_CREDENTIALS
                retval=(*LcmapsRun)(client_name, user_cred_handle, NULL);
#else
                retval=(*LcmapsRun)(user_cred_handle, NULL);
#endif /* ALLOW_EMPTY_CREDENTIALS */
                if (retval)
                {
                    logmsg(LOG_WARNING, "Warning: failed mapping. LCMAPS returned: %d\n", retval);
                    if ((*LcmapsTerm)())
                    {
                        logmsg(LOG_ERR, "LCMAPS Termination failure!");
                        dlclose(lcmaps_handle);
                        return 1;
                    }
                    dlclose(lcmaps_handle);
                    return 1;
                }
                retval=(*LcmapsTerm)();
                if (retval)
                {
                    logmsg(LOG_ERR, "LCMAPS Termination failure!");
                    dlclose(lcmaps_handle);
                    return 1;
                }
                dlclose(lcmaps_handle);
            }
        }
#endif
    /* END USER MAPPING */
    return 0;
}

#endif /* USE_GLOBUS_INTERFACE_TO_LCMAPS  :  Make use of the Globus interface to LCMAPS */
