/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2009-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl> 
 *
 *  2007-2009
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 *  2003-2007
 *     Martijn Steenbakkers <martijn@nikhef.nl>
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 */


/*
 * Author of GT4 Mapping Call Out for LCMAPS:   
 *           Oscar Koeroo <okoeroo@nikhef.nl>
 *
 * See LICENCE file for licence details
 */

/*
 * Globus Gridmap Callout
 * @author Markus Lorch, based on PRIMA module
 *         and Globus skeleton module by Sam Meder
 */

/*
 * Special thanks to:
 *           The UK National Grid Support (NGS) - http://www.grid-support.ac.uk/
 *           Robert Frank    - for supplying the code
 *           Stephen Pickles - for bringing me in contact with Robert Frank (and collegues)
 *
 * for the tips, hints and example code of how they solved this integration work of the Globus GT4 tools utilzing LCAS and LCMAPS in this interface.
 *
 */

#ifndef EXTERN_C_BEGIN
#    ifdef __cplusplus
#        define EXTERN_C_BEGIN extern "C" {
#        define EXTERN_C_END }
#    else
#        define EXTERN_C_BEGIN
#        define EXTERN_C_END
#    endif
#endif


/* General includes */
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>

#include <unistd.h>
#include <pwd.h>
#include <grp.h>





/* LCAS and LCMAPS specific */
#include "lcmaps_gt4_front.h"
#include "utils.h"




EXTERN_C_BEGIN



/* prototypes */
globus_result_t lcmaps_callout( va_list ap);
int run_lcas_authz_and_lcmaps_mapping (gss_ctx_id_t context, char ** local_identity);



int run_lcas_authz_and_lcmaps_mapping (gss_ctx_id_t context, char ** local_identity)
{
    gss_cred_id_t    user_cred_handle = GSS_C_NO_CREDENTIAL;
#ifndef USE_GLOBUS_INTERFACE_TO_LCMAPS    /* When not making use of the Globus interface to LCAS and LCMAPS */
    STACK_OF(X509) * chain            = NULL;
    X509 *           cert             = NULL;
    char *           pem_string       = NULL;
#endif /* When not making use of the Globus interface to LCAS and LCMAPS */

    struct passwd *  pwname           = NULL;

    FILE * flog_lcas   = NULL;
    FILE * flog_lcmaps = NULL;


    /* get user credential from context */
    user_cred_handle = llgt4_get_user_cred_handle(context);

#ifndef USE_GLOBUS_INTERFACE_TO_LCMAPS    /* When not making use of the Globus interface to LCAS and LCMAPS */
    /* Transform the gss_cred_id_t to (STACK_OF(X509) *) and (X509 *) */
    chain = llgt4_cred_to_x509_chain(user_cred_handle);
    if (!chain)
    {
        logmsg(LOG_ERR, "pre-LCAS run: couldn't parse the certificate chain from the GSS layer to a certificate chain\n");
		return 1;
    }
    cert = llgt4_cred_to_x509(user_cred_handle);
    if (!cert)
    {
        logmsg(LOG_ERR, "pre-LCAS run: couldn't parse the certificate from the GSS layer to an individual certificate\n");
		return 1;
    }

    /* Convert Stack of X509 to PEM string */
    if (llgt4_convert_from_stack_of_x509_to_pem (&pem_string, chain))
    {
        logmsg(LOG_ERR, "Client credential conversion from Stack of X509 to PEM string failed\n");
        return 1;
    }

    if (!pem_string)
    {
        logmsg(LOG_ERR, "Couldn't extract the client side credentials.\n");
        return 1;
    }
#endif /* When not making use of the Globus interface to LCAS and LCMAPS */

#if 1
    /* Setup environment for lcas */
    setup_lcas_environment();
    if (getenv ("LCAS_LOG_FILE"))
    {
        flog_lcas = fopen (getenv ("LCAS_LOG_FILE"), "a");
        if (!flog_lcas)
        {
            int errno_fopen = errno;
            char * error_buf = malloc (sizeof(char) * MAX_ERRORBUF_LEN);

            strerror_r (errno_fopen, error_buf, MAX_ERRORBUF_LEN);
            logmsg(LOG_ERR, "Error: failed to open the LCAS log file at the $LCAS_LOG_FILE location: %s. Redirecting LCAS logs Syslog\n", getenv ("LCAS_LOG_FILE"));
            free(error_buf);
        }
    }


#ifndef USE_GLOBUS_INTERFACE_TO_LCMAPS    /* Make use of the Globus interface to LCAS */
    /* Execute LCAS */
    if (run_lcas (pem_string, flog_lcas))
    {
        logmsg(LOG_ERR, "Execution of LCAS failed.\n");
        return 1;
    }
#else  /* USE_GLOBUS_INTERFACE_TO_LCAS  :  Make use of the Globus interface to LCAS */
    /* Execute LCAS with GLOBUS INTERFACE TO LCAS */
    if (run_lcas (user_cred_handle, llgt4_get_client_name(context), flog_lcas))
    {
        logmsg(LOG_ERR, "Execution of LCAS failed.\n");
        return 1;
    }
#endif /* USE_GLOBUS_INTERFACE_TO_LCMAPS  :  Make use of the Globus interface to LCMAPS */

#endif

#if 1
    /* Setup environment for lcmaps*/
    setup_lcmaps_environment();
    if (getenv ("LCMAPS_LOG_FILE"))
    {
        flog_lcmaps = fopen (getenv ("LCMAPS_LOG_FILE"), "a");
        if (!flog_lcmaps)
        {
            int errno_fopen = errno;
            char * error_buf = malloc (sizeof(char) * MAX_ERRORBUF_LEN);

            strerror_r (errno_fopen, error_buf, MAX_ERRORBUF_LEN);
            logmsg(LOG_ERR, "Error: failed to open the LCAS log file at the $LCAS_LOG_FILE location: %s. Redirecting LCAS logs Syslog\n", getenv ("LCAS_LOG_FILE"));
            free(error_buf);
        }
    }


    /* Execute LCMAPS */
#ifndef USE_GLOBUS_INTERFACE_TO_LCMAPS    /* Make use of the Globus interface to LCMAPS */
    if (llgt4_create_jobid() || run_lcmaps(pem_string, flog_lcmaps))
#else  /* USE_GLOBUS_INTERFACE_TO_LCMAPS  :  Make use of the Globus interface to LCMAPS */
    if (llgt4_create_jobid() || run_lcmaps(user_cred_handle, llgt4_get_client_name(context), flog_lcmaps))
#endif /* USE_GLOBUS_INTERFACE_TO_LCMAPS  :  Make use of the Globus interface to LCMAPS */
    {
        logmsg(LOG_ERR, "Execution of LCMAPS failed.\n");
        return 1;
    }
#endif


    /* Am I still Root */
    if ((getuid() != 0) && (geteuid() != 0) && (getgid() != 0) && (getegid() != 0))
    {
        /* Meaning, I'm mapped to an account, so let's tell the calling GT4 tool to which account I was mapped */
        pwname = getpwuid(getuid());
        *local_identity = strdup (pwname->pw_name);

        /* *local_identity = strdup (getpwuid(getuid())->pw_name); */
        logmsg(LOG_NOTICE, "Execution of LCMAPS Succeeded.\n");
        return 0;
    }
    else
    {
        /* Meaning, I'm still root, meaning, no positive LCMAPS outcome */
        *local_identity = NULL;
        logmsg(LOG_ERR, "You are still root after the LCMAPS execution.\n");
        return 1;
    }

    return 0;
}



/* Main function */
globus_result_t lcmaps_callout( va_list ap)
{
    gss_ctx_id_t                        context                 = NULL;
    char *                              service                 = NULL;
    char *                              desired_identity        = NULL;
    char *                              identity_buffer         = NULL;
    unsigned int                        buffer_length           = 0;

    // pointer to a temp buffer with local user name allocated by SAML subroutines
    char *                              local_identity          = NULL;     
    // pointer to temp buffer with primary group name
    char *                              local_group             = NULL;  
    globus_result_t                     result                  = GLOBUS_SUCCESS;
    gss_name_t                          peer                    = NULL;
    gss_buffer_desc                     peer_name_buffer; /* This is a struct */
    gss_name_t                          own                     = NULL;
    gss_buffer_desc                     own_name_buffer; /* This is a struct */
    OM_uint32                           major_status            = 0;
    OM_uint32                           minor_status            = 0;
    int                                 rc                      = 0;
    int                                 initiator               = 0;
    
    //gss_priv_attribute_set              attributes;               
    //gss_id_set                          attribute_types_required;

    //----------------------------------------------------

    context = va_arg(ap, gss_ctx_id_t);
    service = va_arg(ap, char *);
    desired_identity = va_arg(ap, char *);
    identity_buffer = va_arg(ap, char *);
    buffer_length = va_arg(ap, unsigned int);

    //----------------------------------------------------

    /* initialize Syslog */
    openlog ("l_l_gt4", LOG_CONS | LOG_PID | LOG_NDELAY, LOG_LOCAL1);

     // load required modules
    rc = globus_module_activate(GLOBUS_GSI_GSS_ASSIST_MODULE);

    if(rc!=0) {
        logmsg(LOG_ERR, "Error activating Globus GSS ASSIST MODULE");
        result = GLOBUS_FAILURE;
        goto error;
    }

    rc = globus_module_activate(GLOBUS_GSI_GSSAPI_MODULE);

    if(rc!=0) {
        logmsg(LOG_ERR, "Error activating Globus GSSAPI MODULE");
        result = GLOBUS_FAILURE;
        goto error;
    }

    rc = globus_module_activate(GLOBUS_GRIDMAP_CALLOUT_ERROR_MODULE);

    if(rc!=0) {
        logmsg(LOG_ERR, "Error activating Globus GRIDMAP CALLOUT ERROR MODULE");
        result = GLOBUS_FAILURE;
        goto error;
    }



    // find out who is the initiator of this connection (typically the client)
    major_status = gss_inquire_context(&minor_status,
                                       context,
                                       GLOBUS_NULL,
                                       GLOBUS_NULL,
                                       GLOBUS_NULL,
                                       GLOBUS_NULL,
                                       GLOBUS_NULL,
                                       &initiator,
                                       GLOBUS_NULL);

     
    if(GSS_ERROR(major_status))
    {
        GLOBUS_GRIDMAP_CALLOUT_GSS_ERROR(result, major_status, minor_status);
        goto error;
    }

    // get our and the peer's name
    major_status = gss_inquire_context(&minor_status,
                                       context,
                                       initiator ? &own : &peer,
                                       initiator ? &peer : &own,
                                       GLOBUS_NULL,
                                       GLOBUS_NULL,
                                       GLOBUS_NULL,
                                       GLOBUS_NULL,
                                       GLOBUS_NULL);

    if(GSS_ERROR(major_status))
    {
        GLOBUS_GRIDMAP_CALLOUT_GSS_ERROR(result, major_status, minor_status);
        goto error;
    }


    major_status = gss_display_name(&minor_status,
                                    peer,
                                    &peer_name_buffer,
                                    GLOBUS_NULL);

    if(GSS_ERROR(major_status))
    {
        GLOBUS_GRIDMAP_CALLOUT_GSS_ERROR(result, major_status, minor_status);
        gss_release_name(&minor_status, &peer);
        goto error;
    }

    gss_release_name(&minor_status, &peer);


    major_status = gss_display_name(&minor_status,
                                    own,
                                    &own_name_buffer,
                                    GLOBUS_NULL);

    if(GSS_ERROR(major_status))
    {
        GLOBUS_GRIDMAP_CALLOUT_GSS_ERROR(result, major_status, minor_status);
        gss_release_name(&minor_status, &own);
        goto error;
    }

    gss_release_name(&minor_status, &own);


    /*************************
     * Start AuthZ with LCAS *  
     *************************/
    if (run_lcas_authz_and_lcmaps_mapping (context, &local_identity))
    {
        logmsg(LOG_ERR, "Failure in Authorization and Mapping");
        result = GLOBUS_FAILURE;
        goto error;
    }

    logmsg(LOG_NOTICE, "Successfull Authorization and Mapping done");


    //----------------------------------------------------------
    // Copy local identity to Globus Arena and return
    //-----------------------------------------------------------

    if(strlen(local_identity) + 1 > buffer_length)
    {
        GLOBUS_GRIDMAP_CALLOUT_ERROR(
            result,
            GLOBUS_GRIDMAP_CALLOUT_BUFFER_TOO_SMALL,
            ("Local identity length: %d Buffer length: %d\n", strlen(local_identity), buffer_length));

        logmsg(LOG_ERR, "Failed map user. Buffer to small: local identity length: %d Buffer length: %d", strlen(local_identity), buffer_length);
    }
    else
    {
        strcpy(identity_buffer, local_identity);
        logmsg(LOG_NOTICE, "Mapping service \"LCMAPS\" returned local user \"%s\"", identity_buffer);
    }



    gss_release_buffer(&minor_status, &peer_name_buffer);
    gss_release_buffer(&minor_status, &own_name_buffer);


    result = GLOBUS_SUCCESS;

    /* syslog log closing */
    /* closelog(); */

error:
    // free strings that have possibly been allocated
    // if they werent allocated they will be NULL and this does no harm
    free(local_identity);
    free(local_group);


    // ------------------------------
    // Deactivate the Globus modules
    // ------------------------------
    globus_module_deactivate(GLOBUS_GSI_GSSAPI_MODULE);
    globus_module_deactivate(GLOBUS_GSI_GSS_ASSIST_MODULE);
    globus_module_deactivate(GLOBUS_GRIDMAP_CALLOUT_ERROR_MODULE);

    /* syslog log closing */
    /* closelog(); */

    return result;
}



EXTERN_C_END
