/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2009-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl> 
 *
 *  2007-2009
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 *  2003-2007
 *     Martijn Steenbakkers <martijn@nikhef.nl>
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 */

#include "utils.h"


/* static FILE *debug_file[2] = {NULL, NULL}; */


char * llgt4_genfilename(char * prefixp, char * pathp, char * sufixp)
{
    char * newfilename;
    int    prefixl, pathl, sufixl;
    char * prefix,  * path, * sufix;

    prefix = (prefixp) ? prefixp : "";
    path   = (pathp) ? pathp : "";
    sufix  = (sufixp) ? sufixp : "";

    prefixl = strlen(prefix);
    pathl   =  strlen(path);
    sufixl  =  strlen(sufix); 

    newfilename = (char *) calloc(1, (prefixl + pathl + sufixl + 3));
    if (newfilename) 
    {
        if (*path != '/')
        {
            strcat(newfilename, prefix);
            if ((prefixl != 0) && (prefix[prefixl-1] != '/'))
            {
                strcat(newfilename, "/");
            }
        }
        strcat(newfilename, path);
        if ((pathl  != 0) &&
            (sufixl != 0) && 
            (path[pathl-1] != '/') && 
            sufix[0] != '/')
        {
            strcat(newfilename, "/");
        }
        strcat(newfilename, sufix);
    }
    return newfilename;
}

/*
*   FILE* llgt4_register_log(FILE *fh)
*   {
*       int idx = (debug_file[0] == NULL) ? 0 : 1;
*       debug_file[idx] = fh;
*       return debug_file[idx];
*   } 
*
*
*   FILE* llgt4_open_log(const char *filename)
*   {
*       int idx = (debug_file[0] == NULL) ? 0 : 1;
*       debug_file[idx] = fopen(filename, "a");
*       return debug_file[idx];
*   }
*
*   void llgt4_close_log(void)
*   {
*       int idx = (debug_file[1] == NULL) ? 0 : 1;
*       if(debug_file[idx] != NULL)
*       {
*           fclose(debug_file[idx]);
*           debug_file[idx] = NULL;
*       }
*   }
*/


void logmsg (int priority, const char* format, ...)
{
    va_list     args;
    time_t      _time;
    struct tm * _time_s;
    char        _strf_fmt[MAX_TIME_STRING_SIZE];

    _time = time(NULL);
    _time_s = localtime(&_time);
    strftime(_strf_fmt, MAX_TIME_STRING_SIZE, "%G", _time_s);

    va_start(args, format);
    vsyslog(priority, format, args);
    va_end(args);
}



char* llgt4_get_client_name(gss_ctx_id_t context)
{
	int initiator;
	int major_status, minor_status;
	gss_name_t peer;
	gss_buffer_desc peer_name_buffer;
	char *client_name = NULL;

	major_status = gss_inquire_context(&minor_status,
                                       context,
                                       GLOBUS_NULL,
                                       GLOBUS_NULL,
                                       GLOBUS_NULL,
                                       GLOBUS_NULL,
                                       GLOBUS_NULL,
                                       &initiator,
                                       GLOBUS_NULL);
    if(GSS_ERROR(major_status))
    {
//        GLOBUS_GRIDMAP_CALLOUT_GSS_ERROR(result, major_status, minor_status);
//        goto error;
    	return NULL;
    }

	major_status = gss_inquire_context(&minor_status,
                                       context,
                                       initiator ? GLOBUS_NULL : &peer,
                                       initiator ? &peer : GLOBUS_NULL,
                                       GLOBUS_NULL,
                                       GLOBUS_NULL,
                                       GLOBUS_NULL,
                                       GLOBUS_NULL,
                                       GLOBUS_NULL);
    if(GSS_ERROR(major_status))
    {
//        GLOBUS_GRIDMAP_CALLOUT_GSS_ERROR(result, major_status, minor_status);
//        goto error;
		return NULL;
    }
    major_status = gss_display_name(&minor_status,
                                    peer,
                                    &peer_name_buffer,
                                    GLOBUS_NULL);
    if(GSS_ERROR(major_status))
    {
//        GLOBUS_GRIDMAP_CALLOUT_GSS_ERROR(result, major_status, minor_status);
        gss_release_name(&minor_status, &peer);
//        goto error;
		return NULL;
    }
    gss_release_name(&minor_status, &peer);
	if((client_name = (char*)malloc(peer_name_buffer.length + 1)))
	{
		memcpy(client_name, peer_name_buffer.value, peer_name_buffer.length);
		client_name[peer_name_buffer.length] = 0;
	}
	gss_release_buffer(&minor_status, &peer_name_buffer);
	return client_name;
}





/******************************************************************************
Function:       cred_to_x509()
Description:    Return the pointer to X509 structure from gss credential
Parameters:
                cred: globus credential
Returns:        pointer to X509 struct or NULL
******************************************************************************/
/*!
    \fn lcas_cred_to_x509(
        gss_cred_id_t cred
        )
    \brief Return the pointer to X509 structure from gss credential

    This function takes a gss credential as input and returns the corresponding
    X509 structure, which is allocated for this purpose (should be freed)
    
    \param cred the gss credential

    \return a pointer to a X509 struct or NULL
*/
X509 * llgt4_cred_to_x509(
        gss_cred_id_t cred
)
{
    /* Internally a gss_cred_id_t type is a pointer to a gss_cred_id_desc */
    gss_cred_id_desc *       cred_desc = NULL;
    globus_gsi_cred_handle_t gsi_cred_handle;
    X509 * px509=NULL;

    /* cast to gss_cred_id_desc */
    if (cred != GSS_C_NO_CREDENTIAL)
    {
        cred_desc = (gss_cred_id_desc *) cred;
        if (globus_module_activate(GLOBUS_GSI_CREDENTIAL_MODULE) ==GLOBUS_SUCCESS)
        {
            gsi_cred_handle = cred_desc->cred_handle;
            if (globus_gsi_cred_get_cert(gsi_cred_handle, &px509) == GLOBUS_SUCCESS)
            {
                globus_module_deactivate(GLOBUS_GSI_CREDENTIAL_MODULE);
                return px509;
            }
            else
            {
                globus_module_deactivate(GLOBUS_GSI_CREDENTIAL_MODULE);
                return NULL;
            }
        }
        else
        {
            globus_module_deactivate(GLOBUS_GSI_CREDENTIAL_MODULE);
            return NULL;
        }
    }
    else
    {
        return NULL;
    }
}


/******************************************************************************
Function:       cred_to_x509_chain()
Description:    Return the pointer to X509 chain from gss credential
Parameters:
                cred: globus credential
Returns:        pointer to X509 chain or NULL
******************************************************************************/
STACK_OF(X509) * llgt4_cred_to_x509_chain(
        gss_cred_id_t cred
)
{
    /* Internally a gss_cred_id_t type is a pointer to a gss_cred_id_desc */
    gss_cred_id_desc *       cred_desc = NULL;
    globus_gsi_cred_handle_t gsi_cred_handle;
    STACK_OF(X509) * px509_chain=NULL;

    /* cast to gss_cred_id_desc */
    if (cred != GSS_C_NO_CREDENTIAL)
    {
        cred_desc = (gss_cred_id_desc *) cred;
        if (globus_module_activate(GLOBUS_GSI_CREDENTIAL_MODULE) ==GLOBUS_SUCCESS)
        {
            gsi_cred_handle = cred_desc->cred_handle;
            if (globus_gsi_cred_get_cert_chain(gsi_cred_handle, &px509_chain) == GLOBUS_SUCCESS)
            {
                globus_module_deactivate(GLOBUS_GSI_CREDENTIAL_MODULE);
                return px509_chain;
            }
            else
            {
                globus_module_deactivate(GLOBUS_GSI_CREDENTIAL_MODULE);
                return NULL;
            }
        }
        else
        {
            globus_module_deactivate(GLOBUS_GSI_CREDENTIAL_MODULE);
            return NULL;
        }
    }
    else
    {
        return NULL;
    }
}

/******************************************************************************
Function:       convert_from_stack_of_x509_to_pem
Description:    Return the pointer to a PEM string from a X509 chain
Parameters:
                pem: resulting PEM string
                chain: Stack of X509 pointer aka certificate chain
Returns:        
******************************************************************************/
int llgt4_convert_from_stack_of_x509_to_pem (char ** pem, STACK_OF(X509) * chain)
{
    /* TODO: I should make a beter and cleaner interface in LCAS to do this, a newer framework will perform that task in the future */
    BIO * certbio_mem  = NULL;
    X509 * px509       = NULL;
    char * pem_string  = NULL;
    int i = 0;
    int len = 0;

    certbio_mem = BIO_new(BIO_s_mem());

    /* Got chain */
    if (!chain)
    {
        logmsg(LOG_ERR, "%s: Error: Could not get peer chain of certificates.\n", __func__);
        return 1;
    }

    /* Push certificates in chain into the BIO memory stack */
    for (i = 0; i < sk_X509_num(chain); i++)
    {
        int retval = 0;
        char tmp_dn[256]; /* Note... the 256 buffer length is only used for logging output on the spot here. More authoritive information is in the LCAS and LCMAPS logs */

        px509 = sk_X509_value(chain, i);
        X509_NAME_oneline(X509_get_subject_name(px509), tmp_dn, 256);
        logmsg(LOG_DEBUG, "%s: Subject is: %s\n", __func__, tmp_dn);
        retval = PEM_write_bio_X509(certbio_mem, px509);
        logmsg(LOG_DEBUG, "%s: PEM_write_bio_X509 from stack item %d, %s\n", __func__, i, (retval ? "succeeded" : "failed"));
    }


    /* Push BIO memory to character array */
    /* Oscar
     * Warning: The BIO_get_mem_data() will only apply a REFERENCE to 
     * the pem_string, NOT A COPY of the data 
     */
    len = BIO_get_mem_data (certbio_mem, &pem_string);
    pem_string[len]='\0';

    logmsg(LOG_INFO, "%s: PEM string constructed out of certificate chain.\n", __func__);

    *pem = pem_string;

    return 0;
}


int llgt4_grid_x509IsCA(X509 *cert)
{
    int idret;

    /* final argument to X509_check_purpose() is whether to check for CAness */
    idret = X509_check_purpose(cert, X509_PURPOSE_SSL_CLIENT, 1);
    if (idret == 1)
        return 1;
    else if (idret == 0)
        return 0;
    else
    {
        logmsg (LOG_WARNING, "%s: Purpose warning code = %d\n", __func__, idret );
        return 1;
    }
}


void setup_lcas_environment(void)
{
    if (!getenv ("LCAS_LOG_LEVEL"))
        setenv("LCAS_LOG_LEVEL", "5", 1);

    if (!getenv ("LCAS_DEBUG_LEVEL"))
        setenv("LCAS_DEBUG_LEVEL", "5", 1);

    if (!getenv ("LCAS_DB_FILE"))
        setenv("LCAS_DB_FILE", SYSCONFDIR"/lcas/lcas.db", 1);

/* Not required by default, because the setting of the LCMAPS_LOG_FILE will declare to log to file, or use Syslog instead */
#ifdef LCAS_LCMAPS_FORCE_LOG_TO_FILE
    if (!getenv ("LCAS_LOG_FILE"))
        setenv ("LCAS_LOG_FILE",  "/var/log/gt4_lcas_lcmaps.log", 1);
#endif
    
    if (!getenv ("LCAS_DIR"))
        setenv("LCAS_DIR", SYSCONFDIR, 1);

    if (!getenv ("LCAS_MOD_HOME"))
        setenv("LCAS_MOD_HOME", MODULEDIR, 1);
}


void setup_lcmaps_environment(void)
{
    if (!getenv ("LCMAPS_DEBUG_LEVEL"))
       setenv ("LCMAPS_DEBUG_LEVEL", "5", 1);

    if (!getenv ("LCMAPS_LOG_LEVEL"))
        setenv ("LCMAPS_LOG_LEVEL",   "5", 1);

/* Not required by default, because the setting of the LCMAPS_LOG_FILE will declare to log to file, or use Syslog instead */
#ifdef LCAS_LCMAPS_FORCE_LOG_TO_FILE
    if (!getenv ("LCMAPS_LOG_FILE"))
        setenv ("LCMAPS_LOG_FILE",  "/var/log/gt4_lcas_lcmaps.log", 1);
#endif

    if (!getenv ("LCMAPS_DB_FILE"))
        setenv("LCMAPS_DB_FILE", SYSCONFDIR"/lcmaps/lcmaps.db", 1);

    if (!getenv ("LCMAPS_DIR"))
        setenv("LCMAPS_DIR", SYSCONFDIR, 1);

    if (!getenv ("LCMAPS_MOD_HOME"))
        setenv("LCMAPS_MOD_HOME", MODULEDIR, 1);

    if (!getenv ("LCMAPS_POLICY_NAME"))
        setenv ("LCMAPS_POLICY_NAME", "", 1);
}





gss_cred_id_t llgt4_get_user_cred_handle(gss_ctx_id_t context_handle)
{
    return (gss_cred_id_t)((gss_ctx_id_desc *)context_handle)->peer_cred_handle;
}


uid_t llgt4_lcmaps_getuid (char * username)
{
    uid_t res = -1;
    struct passwd * pwd = NULL;    

    pwd = getpwnam(username);
    if (pwd == NULL)
        return res;

    res = pwd->pw_uid;
    free (pwd);

    return res;
}


int llgt4_create_jobid(void)
{
    /*
     * Create the unique Job Repository ID and export it as an environment variable
     * the ID is the same one as for the Job Manager with the addition of microseconds.
     * We moved this part here so that the LCMAPS Job Repository plugin
     * can use this environment variable to uniquely identify the job
     * To have a really unique id we also add the microseconds to the time string
     */
	struct tm  * jr_tmp;
	const char * jr_id_var = "JOB_REPOSITORY_ID";
	char         jr_id[71];
	const char * gk_jm_id_var = "GATEKEEPER_JM_ID";
	char         gatekeeper_jm_id[64];
	const pid_t  gatekeeper_pid = getpid();
	static int   reqnr = 0;

#ifdef HAVE_GETTIMEOFDAY
	{
	struct timeval  jr_clock;

	/* get time in microseconds ! */
	gettimeofday(&jr_clock,NULL);
	jr_tmp = localtime(&(jr_clock.tv_sec));

	snprintf(jr_id, 70, "%04d-%02d-%02d.%02d:%02d:%02d.%06d.%010u.%010u",
	    jr_tmp->tm_year + 1900, jr_tmp->tm_mon + 1, jr_tmp->tm_mday,
	    jr_tmp->tm_hour, jr_tmp->tm_min, jr_tmp->tm_sec, (int)(jr_clock.tv_usec),
	    gatekeeper_pid & 0xFFFFFFFF, reqnr & 0xFFFFFFFF);
	}
#else
	{
	time_t       myclock;

	time(&myclock);
	jr_tmp = localtime(&myclock);
	}

	snprintf(jr_id, 70, "%04d-%02d-%02d.%02d:%02d:%02d.%010u.%010u",
	    jr_tmp->tm_year + 1900, jr_tmp->tm_mon + 1, jr_tmp->tm_mday,
	    jr_tmp->tm_hour, jr_tmp->tm_min, jr_tmp->tm_sec,
	    gatekeeper_pid & 0xFFFFFFFF, reqnr & 0xFFFFFFFF);
#endif
	snprintf(gatekeeper_jm_id, 63, "%04d-%02d-%02d.%02d:%02d:%02d.%010u.%010u",
	    jr_tmp->tm_year + 1900, jr_tmp->tm_mon + 1, jr_tmp->tm_mday,
	    jr_tmp->tm_hour, jr_tmp->tm_min, jr_tmp->tm_sec,
	    gatekeeper_pid & 0xFFFFFFFF, reqnr & 0xFFFFFFFF);


	setenv(jr_id_var, jr_id, 1);
	setenv(gk_jm_id_var, gatekeeper_jm_id, 1);

//	notice3(0, "%s %s (unique id used for Job Repository)", jr_id_var, jr_id);
//	notice(0, "FORMAT:           YYYY-MM-DD.hh:mm:ss.micros.pid.connection");
#ifdef HAVE_GETTIMEOFDAY
//	notice(0, "(Format:          <date>.<time (with microsecs)>.<pid>.<connection counter>)");
#else
//	notice(0, "(Format:          <date>.<time (in seconds)>.<pid>.<connection counter>)");
#endif
	reqnr++;
	return 0;
}
