/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2009-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl> 
 *
 *  2007-2009
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 *  2003-2007
 *     Martijn Steenbakkers <martijn@nikhef.nl>
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 */


#ifndef UTILS_H
  #define UTILS_H


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <syslog.h>
#include <openssl/x509.h>
#include <dlfcn.h>

/* #include "globus_i_gsi_credential.h" */
/* #include "globus_gsi_gss_constants.h" */
/* #include <gssapi.h> */

#include <lcas/lcas.h>
#include <lcmaps/lcmaps.h>


/* For the gss_cred_id_t struct (input) */
/* #include <gssapi.h> */
/* #include "globus_common.h" */
#include "gssapi_openssl.h"
#include <globus_gss_assist.h>
#include <globus_gridmap_callout_error.h>

/* #include "globus_gsi_credential.h" */

/* Need this to dig into the context */
/* #include "globus_gsi_gss_constants.h" */


#define MAX_TIME_STRING_SIZE 256
#define MAX_ERRORBUF_LEN 512

#define FAILED_SERVER 1
#define FAILED_AUTHORIZATION 2



/* Types */
typedef struct __lcas_config
{
        FILE *usrlog_fp;
        char *client_name;
        lcas_request_t lcas_request;
        const char *lcasmod_name;
} lcas_config_t;


typedef struct __lcmaps_config
{
        FILE *usrlog_fp;
        char *client_name;
        lcmaps_request_t lcmaps_request;
        const char *lcmapsmod_name;
} lcmaps_config_t;



/* static FILE *debug_file[2] = {NULL, NULL}; */

char * llgt4_genfilename(char * prefixp, char * pathp, char * sufixp);
/* FILE* llgt4_register_log(FILE *fh); */
/* FILE* llgt4_open_log(const char *filename); */
/* void llgt4_close_log(void); */
void logmsg (int priority, const char* format, ...);
char* llgt4_get_client_name(gss_ctx_id_t context);


int llgt4_convert_from_stack_of_x509_to_pem (char ** pem, STACK_OF(X509) * chain);
X509 * llgt4_cred_to_x509(gss_cred_id_t cred);
STACK_OF(X509) * llgt4_cred_to_x509_chain(gss_cred_id_t cred);
int llgt4_grid_x509IsCA(X509 *cert);

void setup_lcas_environment(void);
void setup_lcmaps_environment(void);
gss_cred_id_t llgt4_get_user_cred_handle(gss_ctx_id_t context_handle);
uid_t llgt4_lcmaps_getuid (char * username);

int llgt4_create_jobid(void);


#ifndef USE_GLOBUS_INTERFACE_TO_LCMAPS
int run_lcas   (char *, FILE *);
int run_lcmaps (char *, FILE *);
#else
int run_lcas (gss_cred_id_t, char *, FILE *);
int run_lcmaps (gss_cred_id_t, char *, FILE *);
#endif /* USE_GLOBUS_INTERFACE_TO_LCMAPS */


#endif /* UTILS_H */
